(*************************************************************)
(*                                                           *)
(*      Congruence of Strong (Late) Bisimilarity w.r.t.      *)
(* the ! (bang) operator (since this result is particularly  *)
(*        cumbersome, we put it in a separate file).         *)
(*                                                           *)
(*************************************************************)
(*                                                           *)
(* File   : bangpack.v                                       *)
(* Author : Ivan Scagnetto (scagnett@dimi.uniud.it),         *)
(*          Dipartimento di Matematica e Informatica,        *)
(*          University of Udine                              *)
(* Date   : July 1998                                        *)
(*                                                           *)
(*************************************************************)
(*                                                           *)
(* This file is part of "Pi-calculus in (Co)Inductive        *)
(* Type Theory", a joint work with Furio Honsell and         *)
(* Marino Miculan (accepted for publication on TCS).         *)
(* If you are interested you can find more information       *)
(* (and a gzipped PostScript version of the article)         *)
(* at the following URL:                                     *)
(* http://www.dimi.uniud.it/~scagnett/pi-calculus.html       *)
(*                                                           *)
(*************************************************************)

(*************************************************************)
(* In the following we will prove the congruence             *)
(* of Strong (Late) Bisimilarity w.r.t. the ! operator.      *)
(* Since the transition rules fBANG and bBANG do not satisfy *)
(* a subformula property, this is a long and difficult task  *)
(* and requires the formulation of a canonical form theory   *)
(* concerning the reducts of the processes                   *)
(* whose head constructor is the bang operator.              *)
(*************************************************************)

Require Export alglaws.

Section structural_congruence3.

(*************************************************************)
(* Auxiliary recursive definitions useful to build processes *)
(* with an unlimited (although finite) number of parallel    *)
(* compositions.                                             *)
(*************************************************************)

Fixpoint cf_f_act [p,q:proc; n:nat]: proc :=
         Case n of
           (par q (bang p))
           [m:nat](par p (cf_f_act p q m))
         end.

Fixpoint cf_b_act [p:proc; q:name->proc; n:nat]: name->proc :=
         Case n of
           ([_:name](par (q _) (bang p)))
           [m:nat]([_:name](par p ((cf_b_act p q m) _)))
         end.

Fixpoint cf_tau1 [p,q:proc; r:name->proc; x:name; n,m:nat]: proc :=
         Case m of
           (par (r x) (cf_f_act p q n))
           [m':nat](par p (cf_tau1 p q r x n m'))
         end.

Fixpoint cf_tau2 [p,q:proc; r:name->proc; x:name; n,m:nat]: proc :=
         Case m of
           (par q ((cf_b_act p r n) x))
           [m':nat](par p (cf_tau2 p q r x n m'))
         end.

Fixpoint cf_tau3 [p:proc; q,r:name->proc; n,m:nat]: proc :=
         Case m of
           (nu [_:name](par (q _) ((cf_b_act p r n) _)))
           [m':nat](par p (cf_tau3 p q r n m'))
         end.

Fixpoint frepl1 [p,q:proc; n:nat]: proc :=
         Case n of
           q
           [m:nat](par p (frepl1 p q m))
         end.

Fixpoint frepl2 [p,q:proc; r:name->proc; x:name; n,m:nat]: proc :=
         Case m of
           (par (r x) (frepl1 p q n))
           [m':nat](par p (frepl2 p q r x n m'))
         end.

Fixpoint frepl3 [p,q:proc; r:name->proc; x:name; n,m:nat]: proc :=
         Case m of
           (par q (frepl1 p (r x) n))
           [m':nat](par p (frepl3 p q r x n m'))
         end.

Fixpoint frepl4 [p:proc; q,r:name->proc; n,m:nat]: proc :=
         Case m of
           (nu [_:name](par (q _) (frepl1 p (r _) n)))
           [m':nat](par p (frepl4 p q r n m'))
         end.

(*******************************************************************)
(* Lemmata about canonical forms of reducts derived from processes *)
(* whose head constructor is the ! operator.                       *)
(*******************************************************************)

Lemma BANG_OUT: (p,q:proc)(x,y:name)
                (ftrans (bang p) (Out x y) q)->
                (Ex [r:proc](Ex [n:nat](q=(cf_f_act p r n) /\ (ftrans p (Out x y) r)))).

Proof.

(Induction q; Intros; Auto).

Inversion_clear H; Inversion_clear H0.

Inversion_clear H0; Inversion_clear H1.

Inversion_clear H0; Inversion_clear H1.

(Inversion_clear H1; Inversion H2).
Exists p0; Exists O; Simpl; Split; [Trivial | Assumption].
Rewrite <-H6 in H5; Rewrite <-H6; Elim (H0 x y H5); Intros.
Inversion_clear H8.
Inversion_clear H9.
Exists x0; Exists (S x1); Split; [Simpl; Rewrite H8; Trivial | Assumption].

(Inversion_clear H1; Inversion_clear H2).

(Inversion_clear H0; Inversion_clear H1).

(Inversion_clear H0; Inversion_clear H1).

(Inversion_clear H0; Inversion_clear H1).

(Inversion_clear H0; Inversion_clear H1).

(Inversion_clear H0; Inversion_clear H1).

Qed.


Lemma PRE_BANG_BTR: (p,q:proc)(r:name->proc)(a:b_act)(x:name)
                    (notin x (nu r))->(notin x p)->(q=(r x))->
                    (btrans (bang p) a r)->
                    (Ex [s:name->proc](Ex [n:nat](r=(cf_b_act p s n) /\ (btrans p a s)))).

Proof.

(Induction q; Intros; Auto).

Cut (r=[_:name]nil); [Intro | Apply proc_ext with x; Auto].
Rewrite H3 in H2; Inversion_clear H2.
Inversion H4; [Absurd (par (p0 x) (bang p))=nil; [Discriminate | Change (([_:name](par (p0 _) (bang p))) x)=(([_:name]nil) x); Rewrite H8; Trivial] | Absurd (par p (p0 x))=nil; [Discriminate | Change (([_:name](par p (p0 _))) x)=(([_:name]nil) x); Rewrite H8; Trivial]].
Apply notin_nu; Intros; Apply notin_nil.

Elim (proc_exp p0 x); Intros.
Inversion_clear H4.
Rewrite H6 in H2.
Cut r=([_:name](bang (x0 _))); [Intro | Apply proc_ext with x; Auto].
Rewrite H4 in H3; Inversion_clear H3.
Inversion H7; [Absurd (par (p3 x) (bang p))=(bang (x0 x)); [Discriminate | Change (([_:name](par (p3 _) (bang p))) x)=(([_:name](bang (x0 _))) x); Rewrite H11; Trivial] | Absurd (par p (p3 x))=(bang (x0 x)); [Discriminate | Change (([_:name](par p (p3 _))) x)=(([_:name](bang (x0 _))) x); Rewrite H11; Trivial]].
Apply notin_nu; Intros; Inversion_clear H5; Apply notin_bang; Auto.

Elim (proc_exp p0 x); Intros.
Inversion_clear H4.
Rewrite H6 in H2.
Cut r=([_:name](tau_pref (x0 _))); [Intro | Apply proc_ext with x; Auto].
Rewrite H4 in H3; Inversion_clear H3.
Inversion H7; [Absurd (par (p3 x) (bang p))=(tau_pref (x0 x)); [Discriminate | Change (([_:name](par (p3 _) (bang p))) x)=(([_:name](tau_pref (x0 _))) x); Rewrite H11; Trivial] | Absurd (par p (p3 x))=(tau_pref (x0 x)); [Discriminate | Change (([_:name](par p (p3 _))) x)=(([_:name](tau_pref (x0 _))) x); Rewrite H11; Trivial]].
Apply notin_nu; Intros; Inversion_clear H5; Apply notin_tau; Auto.

Inversion_clear H4.
Inversion H5.
Exists p4; Exists O; Simpl; Split; [Trivial | Assumption].
Cut p1=(p4 x); [Intro | Rewrite <- H8 in H3; Inversion_clear H3; Trivial].
Cut (notin x (nu p4)); [Intro | Rewrite <- H8 in H1; Inversion_clear H1].
Cut (EX s:name->proc | (EX n:nat | p4=(cf_b_act p s n)/\(btrans p a s))); [Intro | Apply H0 with x; Auto].
Inversion_clear H12.
Inversion_clear H13.
Inversion_clear H12.
Exists x0; Exists (S x1); Split; [Simpl; Rewrite <- H13; Trivial | Assumption].
Apply notin_nu; Intros; Cut (notin x (par p (p4 z))); [Intro | Auto].
Inversion_clear H12; Assumption.

Elim (proc_exp p0 x); Intros; Elim (proc_exp p1 x); Intros.
Inversion_clear H5; Inversion_clear H6.
Rewrite H8 in H3; Rewrite H9 in H3.
Cut r=([_:name](sum (x0 _) (x1 _))); [Intro | Apply proc_ext with x; Auto].
Rewrite H6 in H4; Inversion_clear H4.
Inversion H10; [Absurd (par (p4 x) (bang p))=(sum (x0 x) (x1 x)); [Discriminate | Change (([_:name](par (p4 _) (bang p))) x)=(([_:name](sum (x0 _) (x1 _))) x); Rewrite H14; Trivial] | Absurd (par p (p4 x))=(sum (x0 x) (x1 x)); [Discriminate | Change (([_:name](par p (p4 _))) x)=(([_:name](sum (x0 _) (x1 _))) x); Rewrite H14; Trivial]].
Apply notin_nu; Intros; Inversion_clear H7; Inversion_clear H5; Apply notin_sum; Auto.

Elim (ho_proc_exp p0 x); Intros.
Inversion_clear H4.
Rewrite H6 in H2.
Cut r=([_:name](nu (x0 _))); [Intro | Apply proc_ext with x; Auto].
Rewrite H4 in H3; Inversion_clear H3.
Inversion H7; [Absurd (par (p3 x) (bang p))=(nu (x0 x)); [Discriminate | Change (([_:name](par (p3 _) (bang p))) x)=(([_:name](nu (x0 _))) x); Rewrite H11; Trivial] | Absurd (par p (p3 x))=(nu (x0 x)); [Discriminate | Change (([_:name](par p (p3 _))) x)=(([_:name](nu (x0 _))) x); Rewrite H11; Trivial]].

Elim (proc_exp p0 x); Intros.
Inversion_clear H4.
Rewrite H6 in H2.
Elim (LEM_name n x); Intros; Elim (LEM_name n0 x); Intros.
Rewrite H4 in H2; Rewrite H7 in H2.
Cut r=([_:name](match _ _ (x0 _))); [Intro | Apply proc_ext with x; Auto].
Rewrite H8 in H3; Inversion_clear H3.
Inversion H9; [Absurd (par (p3 x) (bang p))=(match x x (x0 x)); [Discriminate | Change (([_:name](par (p3 _) (bang p))) x)=(([_:name](match _ _ (x0 _))) x); Rewrite H13; Trivial] | Absurd (par p (p3 x))=(match x x (x0 x)); [Discriminate | Change (([_:name](par p (p3 _))) x)=(([_:name](match _ _ (x0 _))) x); Rewrite H13; Trivial]].
Apply notin_nu; Intros; Inversion_clear H5; Apply notin_match; Auto.

Rewrite H4 in H2.
Cut r=([_:name](match _ n0 (x0 _))); [Intro | Apply proc_ext with x; Auto].
Rewrite H8 in H3; Inversion_clear H3.
Inversion H9; [Absurd (par (p3 x) (bang p))=(match x n0 (x0 x)); [Discriminate | Change (([_:name](par (p3 _) (bang p))) x)=(([_:name](match _ n0 (x0 _))) x); Rewrite H13; Trivial] | Absurd (par p (p3 x))=(match x n0 (x0 x)); [Discriminate | Change (([_:name](par p (p3 _))) x)=(([_:name](match _ n0 (x0 _))) x); Rewrite H13; Trivial]].
Apply notin_nu; Intros; Inversion_clear H5; Apply notin_match; Auto.

Rewrite H7 in H2.
Cut r=([_:name](match n _ (x0 _))); [Intro | Apply proc_ext with x; Auto].
Rewrite H8 in H3; Inversion_clear H3.
Inversion H9; [Absurd (par (p3 x) (bang p))=(match n x (x0 x)); [Discriminate | Change (([_:name](par (p3 _) (bang p))) x)=(([_:name](match n _ (x0 _))) x); Rewrite H13; Trivial] | Absurd (par p (p3 x))=(match n x (x0 x)); [Discriminate | Change (([_:name](par p (p3 _))) x)=(([_:name](match n _ (x0 _))) x); Rewrite H13; Trivial]].
Apply notin_nu; Intros; Inversion_clear H5; Apply notin_match; Auto.

Cut r=([_:name](match n n0 (x0 _))); [Intro | Apply proc_ext with x; Auto].
Rewrite H8 in H3; Inversion_clear H3.
Inversion H9; [Absurd (par (p3 x) (bang p))=(match n n0 (x0 x)); [Discriminate | Change (([_:name](par (p3 _) (bang p))) x)=(([_:name](match n n0 (x0 _))) x); Rewrite H13; Trivial] | Absurd (par p (p3 x))=(match n n0 (x0 x)); [Discriminate | Change (([_:name](par p (p3 _))) x)=(([_:name](match n n0 (x0 _))) x); Rewrite H13; Trivial]].
Apply notin_nu; Intros; Inversion_clear H5; Apply notin_match; Auto.

Elim (proc_exp p0 x); Intros.
Inversion_clear H4.
Rewrite H6 in H2.
Elim (LEM_name n x); Intros; Elim (LEM_name n0 x); Intros.
Rewrite H4 in H2; Rewrite H7 in H2.
Cut r=([_:name](mismatch _ _ (x0 _))); [Intro | Apply proc_ext with x; Auto].
Rewrite H8 in H3; Inversion_clear H3.
Inversion H9; [Absurd (par (p3 x) (bang p))=(mismatch x x (x0 x)); [Discriminate | Change (([_:name](par (p3 _) (bang p))) x)=(([_:name](mismatch _ _ (x0 _))) x); Rewrite H13; Trivial] | Absurd (par p (p3 x))=(mismatch x x (x0 x)); [Discriminate | Change (([_:name](par p (p3 _))) x)=(([_:name](mismatch _ _ (x0 _))) x); Rewrite H13; Trivial]].
Apply notin_nu; Intros; Inversion_clear H5; Apply notin_mismatch; Auto.

Rewrite H4 in H2.
Cut r=([_:name](mismatch _ n0 (x0 _))); [Intro | Apply proc_ext with x; Auto].
Rewrite H8 in H3; Inversion_clear H3.
Inversion H9; [Absurd (par (p3 x) (bang p))=(mismatch x n0 (x0 x)); [Discriminate | Change (([_:name](par (p3 _) (bang p))) x)=(([_:name](mismatch _ n0 (x0 _))) x); Rewrite H13; Trivial] | Absurd (par p (p3 x))=(mismatch x n0 (x0 x)); [Discriminate | Change (([_:name](par p (p3 _))) x)=(([_:name](mismatch _ n0 (x0 _))) x); Rewrite H13; Trivial]].
Apply notin_nu; Intros; Inversion_clear H5; Apply notin_mismatch; Auto.

Rewrite H7 in H2.
Cut r=([_:name](mismatch n _ (x0 _))); [Intro | Apply proc_ext with x; Auto].
Rewrite H8 in H3; Inversion_clear H3.
Inversion H9; [Absurd (par (p3 x) (bang p))=(mismatch n x (x0 x)); [Discriminate | Change (([_:name](par (p3 _) (bang p))) x)=(([_:name](mismatch n _ (x0 _))) x); Rewrite H13; Trivial] | Absurd (par p (p3 x))=(mismatch n x (x0 x)); [Discriminate | Change (([_:name](par p (p3 _))) x)=(([_:name](mismatch n _ (x0 _))) x); Rewrite H13; Trivial]].
Apply notin_nu; Intros; Inversion_clear H5; Apply notin_mismatch; Auto.

Cut r=([_:name](mismatch n n0 (x0 _))); [Intro | Apply proc_ext with x; Auto].
Rewrite H8 in H3; Inversion_clear H3.
Inversion H9; [Absurd (par (p3 x) (bang p))=(mismatch n n0 (x0 x)); [Discriminate | Change (([_:name](par (p3 _) (bang p))) x)=(([_:name](mismatch n n0 (x0 _))) x); Rewrite H13; Trivial] | Absurd (par p (p3 x))=(mismatch n n0 (x0 x)); [Discriminate | Change (([_:name](par p (p3 _))) x)=(([_:name](mismatch n n0 (x0 _))) x); Rewrite H13; Trivial]].
Apply notin_nu; Intros; Inversion_clear H5; Apply notin_mismatch; Auto.

Elim (ho_proc_exp p0 x); Intros.
Inversion_clear H4.
Rewrite H6 in H2.
Elim (LEM_name n x); Intros.
Rewrite H4 in H2.
Cut r=([_:name](in_pref _ (x0 _))); [Intro | Apply proc_ext with x; Auto].
Rewrite H7 in H3; Inversion_clear H3.
Inversion H8; [Absurd (par (p3 x) (bang p))=(in_pref x (x0 x)); [Discriminate | Change (([_:name](par (p3 _) (bang p))) x)=(([_:name](in_pref _ (x0 _))) x); Rewrite H12; Trivial] | Absurd (par p (p3 x))=(in_pref x (x0 x)); [Discriminate | Change (([_:name](par p (p3 _))) x)=(([_:name](in_pref _ (x0 _))) x); Rewrite H12; Trivial]].
Apply notin_nu; Intros; Inversion_clear H5.
Cut (notin x (nu (x0 z))); [Intro | Auto].
Inversion_clear H5; Apply notin_in; Intros; Auto.

Cut r=([_:name](in_pref n (x0 _))); [Intro | Apply proc_ext with x; Auto].
Rewrite H7 in H3; Inversion_clear H3.
Inversion H8; [Absurd (par (p3 x) (bang p))=(in_pref n (x0 x)); [Discriminate | Change (([_:name](par (p3 _) (bang p))) x)=(([_:name](in_pref n (x0 _))) x); Rewrite H12; Trivial] | Absurd (par p (p3 x))=(in_pref n (x0 x)); [Discriminate | Change (([_:name](par p (p3 _))) x)=(([_:name](in_pref n (x0 _))) x); Rewrite H12; Trivial]].
Apply notin_nu; Intros; Inversion_clear H5.
Cut (notin x (nu (x0 z))); [Intro | Auto].
Inversion_clear H5; Apply notin_in; Intros; Auto.

Elim (proc_exp p0 x); Intros.
Inversion_clear H4.
Rewrite H6 in H2.
Elim (LEM_name n x); Intros; Elim (LEM_name n0 x); Intros.
Rewrite H4 in H2; Rewrite H7 in H2.
Cut r=([_:name](out_pref _ _ (x0 _))); [Intro | Apply proc_ext with x; Auto].
Rewrite H8 in H3; Inversion_clear H3.
Inversion H9; [Absurd (par (p3 x) (bang p))=(out_pref x x (x0 x)); [Discriminate | Change (([_:name](par (p3 _) (bang p))) x)=(([_:name](out_pref _ _ (x0 _))) x); Rewrite H13; Trivial] | Absurd (par p (p3 x))=(out_pref x x (x0 x)); [Discriminate | Change (([_:name](par p (p3 _))) x)=(([_:name](out_pref _ _ (x0 _))) x); Rewrite H13; Trivial]].
Apply notin_nu; Intros; Inversion_clear H5; Apply notin_out; Auto.

Rewrite H4 in H2.
Cut r=([_:name](out_pref _ n0 (x0 _))); [Intro | Apply proc_ext with x; Auto].
Rewrite H8 in H3; Inversion_clear H3.
Inversion H9; [Absurd (par (p3 x) (bang p))=(out_pref x n0 (x0 x)); [Discriminate | Change (([_:name](par (p3 _) (bang p))) x)=(([_:name](out_pref _ n0 (x0 _))) x); Rewrite H13; Trivial] | Absurd (par p (p3 x))=(out_pref x n0 (x0 x)); [Discriminate | Change (([_:name](par p (p3 _))) x)=(([_:name](out_pref _ n0 (x0 _))) x); Rewrite H13; Trivial]].
Apply notin_nu; Intros; Inversion_clear H5; Apply notin_out; Auto.

Rewrite H7 in H2.
Cut r=([_:name](out_pref n _ (x0 _))); [Intro | Apply proc_ext with x; Auto].
Rewrite H8 in H3; Inversion_clear H3.
Inversion H9; [Absurd (par (p3 x) (bang p))=(out_pref n x (x0 x)); [Discriminate | Change (([_:name](par (p3 _) (bang p))) x)=(([_:name](out_pref n _ (x0 _))) x); Rewrite H13; Trivial] | Absurd (par p (p3 x))=(out_pref n x (x0 x)); [Discriminate | Change (([_:name](par p (p3 _))) x)=(([_:name](out_pref n _ (x0 _))) x); Rewrite H13; Trivial]].
Apply notin_nu; Intros; Inversion_clear H5; Apply notin_out; Auto.

Cut r=([_:name](out_pref n n0 (x0 _))); [Intro | Apply proc_ext with x; Auto].
Rewrite H8 in H3; Inversion_clear H3.
Inversion H9; [Absurd (par (p3 x) (bang p))=(out_pref n n0 (x0 x)); [Discriminate | Change (([_:name](par (p3 _) (bang p))) x)=(([_:name](out_pref n n0 (x0 _))) x); Rewrite H13; Trivial] | Absurd (par p (p3 x))=(out_pref n n0 (x0 x)); [Discriminate | Change (([_:name](par p (p3 _))) x)=(([_:name](out_pref n n0 (x0 _))) x); Rewrite H13; Trivial]].
Apply notin_nu; Intros; Inversion_clear H5; Apply notin_out; Auto.

Qed.

Lemma BANG_BTR: (p:proc)(q:name->proc)(a:b_act)
                (btrans (bang p) a q)->
                (Ex [r:name->proc](Ex [n:nat](q=(cf_b_act p r n) /\ (btrans p a r)))).

Proof.

Intros; Elim (unsat (par p (nu q)) empty); Intros; Inversion_clear H0; Inversion_clear H1; Clear H2; Apply PRE_BANG_BTR with (q x) x; Auto.

Qed.

Lemma BANG_TAU: (p,q:proc)
                (ftrans (bang p) tau q)->
                (Ex [r:proc](Ex [n:nat]q=(cf_f_act p r n) /\ (ftrans p tau r))) \/
                (Ex [r:proc]
                  (Ex [s:name->proc]
                    (Ex [x:name]
                      (Ex [y:name]
                        (Ex [n:nat]
                          (Ex [m:nat]q=(cf_tau1 p r s y n m) /\
                                     (ftrans p (Out x y) r) /\
                                     (btrans p (In x) s)
                          )
                        )
                      )
                    )
                  )
                ) \/
                (Ex [r:proc]
                  (Ex [s:name->proc]
                    (Ex [x:name]
                      (Ex [y:name]
                        (Ex [n:nat]
                          (Ex [m:nat]q=(cf_tau2 p r s y n m) /\
                                     (ftrans p (Out x y) r) /\
                                     (btrans p (In x) s)
                          )
                        )
                      )
                    )
                  )
                ) \/
                (Ex [r:name->proc]
                  (Ex [s:name->proc]
                    (Ex [x:name]
                      (Ex [n:nat]
                        (Ex [m:nat]q=(cf_tau3 p r s n m)  /\
                                   (btrans p (In x) r) /\
                                   (btrans p (bOut x) s)
                        )
                      )
                    )
                  )
                ) \/
                (Ex [r:name->proc]
                  (Ex [s:name->proc]
                    (Ex [x:name]
                      (Ex [n:nat]
                        (Ex [m:nat]q=(cf_tau3 p s r n m)  /\
                                   (btrans p (In x) r) /\
                                   (btrans p (bOut x) s)
                        )
                      )
                    )
                  )
                ).

Proof.

(Induction q; Intros; Auto).

Inversion_clear H; Inversion_clear H0.

Inversion_clear H0; Inversion_clear H1.

Inversion_clear H0; Inversion_clear H1.

Inversion_clear H1; Inversion H2.
Left; Exists p0; Exists O; Split; [Simpl; Trivial | Assumption].
Rewrite <- H6 in H5; Rewrite <- H6; Elim (H0 H5); Intros.
Inversion_clear H8.
Inversion_clear H9.
Inversion_clear H8.
Left; Exists x; Exists (S x0); Split; [Simpl; Rewrite H9; Trivial | Assumption].
Elim H8; Intros.
Inversion_clear H9.
Inversion_clear H10.
Inversion_clear H9.
Inversion_clear H10.
Inversion_clear H9.
Inversion_clear H10.
Inversion_clear H9.
Right; Left; Exists x; Exists x0; Exists x1; Exists x2; Exists x3; Exists (S x4); Split; [Simpl; Rewrite H10; Trivial | Assumption].
Elim H9; Intros.
Inversion_clear H10.
Inversion_clear H11.
Inversion_clear H10.
Inversion_clear H11.
Inversion_clear H10.
Inversion_clear H11.
Inversion_clear H10.
Right; Right; Left; Exists x; Exists x0; Exists x1; Exists x2; Exists x3; Exists (S x4); Split; [Simpl; Rewrite H11; Trivial | Assumption].
Elim H10; Intros.
Inversion_clear H11.
Inversion_clear H12.
Inversion_clear H11.
Inversion_clear H12.
Inversion_clear H11.
Inversion_clear H12.
Right; Right; Right; Left; Exists x; Exists x0; Exists x1; Exists x2; Exists (S x3); Split; [Simpl; Rewrite H11; Trivial | Assumption].
Inversion_clear H11.
Inversion_clear H12.
Inversion_clear H11.
Inversion_clear H12.
Inversion_clear H11.
Inversion_clear H12.
Right; Right; Right; Right; Exists x; Exists x0; Exists x1; Exists x2; Exists (S x3); Split; [Simpl; Rewrite H11; Trivial | Assumption].

Elim (BANG_OUT p p1 x y H8); Intros.
Inversion_clear H9.
Inversion_clear H10.
Right; Left; Exists x0; Exists q1; Exists x; Exists y; Exists x1; Exists O; Split; [Simpl; Rewrite H9; Trivial | Split; Assumption].

Elim (BANG_BTR p q2 (In x) H8); Intros.
Inversion_clear H9.
Inversion_clear H10.
Right; Right; Left; Exists p0; Exists x0; Exists x; Exists y; Exists x1; Exists O; Split; [Simpl; Rewrite H9; Trivial | Split; Assumption].

Inversion_clear H1; Inversion_clear H2.

Inversion_clear H0; Inversion_clear H1.
Elim (BANG_BTR p q2 (bOut x) H2); Intros.
Inversion_clear H1.
Inversion_clear H3.
Right; Right; Right; Left; Exists q1; Exists x0; Exists x; Exists x1; Exists O; Split; [Simpl; Rewrite H1; Trivial | Split; Assumption].
Elim (BANG_BTR p q2 (In x) H2); Intros.
Inversion_clear H1.
Inversion_clear H3.
Right; Right; Right; Right; Exists x0; Exists q1; Exists x; Exists x1; Exists O; Split; [Simpl; Rewrite H1; Trivial | Split; Assumption].

Inversion_clear H0; Inversion_clear H1.

Inversion_clear H0; Inversion_clear H1.

Inversion_clear H0; Inversion_clear H1.

Inversion_clear H0; Inversion_clear H1.

Qed.

(*******************************************************)
(* Technical lemmata needed in order to isolate        *)
(* the recursive component from the canonical reducts. *)
(*******************************************************)

Lemma SB_CF_F_ACT: (p,q:proc)(n:nat)(StBisim (cf_f_act p q n) (par (frepl1 p q n) (bang p))).

Proof.

Induction n; Intros.

Simpl; Apply REF.

Simpl; Apply TRANS with (par p (par (frepl1 p q n0) (bang p))); [Apply PAR_S2; [Apply REF | Assumption] | Apply SYM; Apply ASS_PAR].

Qed.

Lemma SB_CF_B_ACT: (p:proc)(q:name->proc)(n:nat)(x:name)(StBisim (cf_b_act p q n x) (par (frepl1 p (q x) n) (bang p))).

Proof.

Induction n; Intros.

Simpl; Apply REF.

Simpl; Apply TRANS with (par p (par (frepl1 p (q x) n0) (bang p))); [Apply PAR_S2; [Apply REF | Auto] | Apply SYM; Apply ASS_PAR].

Qed.

Lemma SB_CF_TAU1: (p,q:proc)(r:name->proc)(y:name)(n,m:nat)(StBisim (cf_tau1 p q r y n m) (par (frepl2 p q r y n m) (bang p))).

Proof.

Induction m; Intros.

Simpl; Apply TRANS with (par (r y) (par (frepl1 p q n) (bang p))); [Apply PAR_S2; [Apply REF | Apply SB_CF_F_ACT] | Apply SYM; Apply ASS_PAR].

Simpl.

Simpl; Apply TRANS with (par p (par (frepl2 p q r y n n0) (bang p))); [Apply PAR_S2; [Apply REF | Assumption] | Apply SYM; Apply ASS_PAR].

Qed.

Lemma SB_CF_TAU2: (p,q:proc)(r:name->proc)(y:name)(n,m:nat)(StBisim (cf_tau2 p q r y n m) (par (frepl3 p q r y n m) (bang p))).

Proof.

Induction m; Intros.

Simpl; Apply TRANS with (par q (par (frepl1 p (r y) n) (bang p))); [Apply PAR_S2; [Apply REF | Apply SB_CF_B_ACT] | Apply SYM; Apply ASS_PAR].

Simpl; Apply TRANS with (par p (par (frepl3 p q r y n n0) (bang p))); [Apply PAR_S2; [Apply REF | Assumption] | Apply SYM; Apply ASS_PAR].

Qed.

Lemma SB_CF_TAU3: (p:proc)(q,r:name->proc)(y:name)(n,m:nat)(StBisim (cf_tau3 p q r n m) (par (frepl4 p q r n m) (bang p))).

Proof.

Induction m; Intros.

Simpl; Apply TRANS with (nu [_:name](par (q _) (par (frepl1 p (r _) n) (bang p)))); [Apply NU_S; Intros; Apply PAR_S2; [Apply REF | Apply SB_CF_B_ACT] | Apply TRANS with (nu [_:name](par (par (q _) (frepl1 p (r _) n)) (bang p))); [Apply NU_S; Intros; Apply SYM; Apply ASS_PAR | Change (StBisim (nu [_:name](par ([u:name] (par (q u) (frepl1 p (r u) n)) _) (bang p))) (par (nu [_:name](par (q _) (frepl1 p (r _) n))) (bang p))); Apply NU_EXTR]].

Simpl; Apply TRANS with (par p (par (frepl4 p q r n n0) (bang p))); [Apply PAR_S2; [Apply REF | Assumption] | Apply SYM; Apply ASS_PAR].

Qed.

(**********************************************************)
(* Technical lemmata about equivalences between processes *)
(* with an unlimited number of parallel compositions.     *)
(**********************************************************)

Lemma F_REPL1: (p,q,r,s:proc)(n:nat)(StBisim p q)->(StBisim r s)->(StBisim (frepl1 p r n) (frepl1 q s n)).

Proof.

Induction n; Intros.
Simpl; Assumption.
Simpl; Apply PAR_S2; [Assumption | Apply H; Assumption].

Qed.

Lemma F_REPL2: (p,q,r,s:proc)(t,u:name->proc)(x:name)(n,m:nat)(StBisim p q)->(StBisim r s)->
               (StBisim (t x) (u x))->(StBisim (frepl2 p r t x n m) (frepl2 q s u x n m)).

Proof.

Induction m; Intros.

Simpl; Apply PAR_S2; [Assumption | Apply F_REPL1; Assumption].

Simpl; Apply PAR_S2; Auto.

Qed.

Lemma F_REPL3: (p,q,r,s:proc)(t,u:name->proc)(x:name)(n,m:nat)(StBisim p q)->(StBisim r s)->
               (StBisim (t x) (u x))->(StBisim (frepl3 p r t x n m) (frepl3 q s u x n m)).

Proof.

Induction m; Intros.

Simpl; Apply PAR_S2; [Assumption | Apply F_REPL1; Assumption].

Simpl; Apply PAR_S2; Auto.

Qed.

Lemma FREPL1_NOTIN: (p,q:proc)(n:nat)(x:name)(notin x (frepl1 p q n))->(notin x q).

Proof.

Induction n; Intros.

Simpl in H; Assumption.

Simpl in H0; Inversion_clear H0; Apply H; Assumption.

Qed.

Lemma F_REPL4: (p,q:proc)(r,s,t,u:name->proc)(n,m:nat)
               (StBisim p q)->
               ((x:name)(notin x (nu r))->(notin x (nu s))->(StBisim (r x) (s x)))->
               ((x:name)(notin x (nu t))->(notin x (nu u))->(StBisim (t x) (u x)))->
               (StBisim (frepl4 p r t n m) (frepl4 q s u n m)).

Proof.

Induction m; Intros.

Simpl; Apply NU_S; Intros; Apply PAR_S2.
Apply H0; Inversion_clear H2; Apply notin_nu; Intros.
Cut (notin x (par (r z) (frepl1 p (t z) n))); [Intro | Auto].
Inversion_clear H5; Assumption.
Inversion_clear H3; Cut (notin x (par (s z) (frepl1 q (u z) n))); [Intro | Auto].
Inversion_clear H3; Assumption.
Apply F_REPL1.
Assumption.
Apply H1; Inversion_clear H2; Apply notin_nu; Intros.
Cut (notin x (par (r z) (frepl1 p (t z) n))); [Intro | Auto].
Inversion_clear H5.
Apply FREPL1_NOTIN with p n; Assumption.
Inversion_clear H3.
Cut (notin x (par (s z) (frepl1 q (u z) n))); [Intro | Auto].
Inversion_clear H3.
Apply FREPL1_NOTIN with q n; Assumption.

Simpl; Apply PAR_S2; [Assumption | Apply H; Auto].

Qed.

(*******************************************************************)
(* Other technical lemmata needed in order to prove the congruence *)
(* of Strong (Late) Bisimilarity w.r.t. the ! operator.            *)
(*******************************************************************)

Lemma CF_B_ACT_NOTIN: (p:proc)(q:name->proc)(n:nat)(x:name)(notin x (nu (cf_b_act p q n)))->(notin x (nu q)).

Proof.

Induction n; Intros.

Simpl in H; Inversion_clear H; Apply notin_nu; Intros.
Cut (notin x (par (q z) (bang p))); [Intro | Auto].
Inversion_clear H1; Assumption.

Simpl in H0; Inversion_clear H0; Apply H; Auto.
Apply notin_nu; Intros.
Cut (notin x (par p (cf_b_act p q n0 z))); [Intro | Auto].
Inversion_clear H2; Assumption.

Qed.

Lemma BANG_FTR_AUX: (p,q:proc)(a:f_act)(n:nat)(ftrans p a q)->(ftrans (bang p) a (cf_f_act p q n)).

Proof.

Induction n; Intros.
Simpl; Apply fBANG; Apply fPAR1; Assumption.
Simpl; Apply fBANG; Apply fPAR2; Apply H; Auto.

Qed.

Lemma BANG_BTR_AUX: (p:proc)(q:name->proc)(a:b_act)(n:nat)(btrans p a q)->(btrans (bang p) a (cf_b_act p q n)).

Proof.

Induction n; Intros.
Simpl; Apply bBANG; Apply bPAR1; Assumption.
Simpl; Apply bBANG; Apply bPAR2; Apply H; Auto.

Qed.

Lemma BANG_TAU_AUX1: (p,q:proc)(r:name->proc)(x,y:name)(n,m:nat)
                     (ftrans p (Out x y) q)->(btrans p (In x) r)->
                     (ftrans (bang p) tau (cf_tau1 p q r y n m)).

Proof.

Induction m; Intros.
Simpl; Apply fBANG; Apply COM1 with x; [Assumption | Apply BANG_FTR_AUX; Assumption].
Simpl; Apply fBANG; Apply fPAR2; Apply H; Assumption.

Qed.

Lemma BANG_TAU_AUX2: (p,q:proc)(r:name->proc)(x,y:name)(n,m:nat)
                     (ftrans p (Out x y) q)->(btrans p (In x) r)->
                     (ftrans (bang p) tau (cf_tau2 p q r y n m)).

Proof.

Induction m; Intros.
Simpl; Apply fBANG; Apply COM2 with x; [Assumption | Apply BANG_BTR_AUX; Assumption].
Simpl; Apply fBANG; Apply fPAR2; Apply H; Assumption.

Qed.

Lemma BANG_TAU_AUX3: (p:proc)(q,r:name->proc)(x:name)(n,m:nat)
                     (btrans p (In x) q)->(btrans p (bOut x) r)->
                     (ftrans (bang p) tau (cf_tau3 p q r n m)).

Proof.

Induction m; Intros.
Simpl; Apply fBANG; Apply CLOSE1 with x; [Assumption | Apply BANG_BTR_AUX; Assumption].
Simpl; Apply fBANG; Apply fPAR2; Apply H; Assumption.

Qed.

Lemma BANG_TAU_AUX4: (p:proc)(q,r:name->proc)(x:name)(n,m:nat)
                     (btrans p (bOut x) q)->(btrans p (In x) r)->
                     (ftrans (bang p) tau (cf_tau3 p q r n m)).

Proof.

Induction m; Intros.
Simpl; Apply fBANG; Apply CLOSE2 with x; [Assumption | Apply BANG_BTR_AUX; Assumption].
Simpl; Apply fBANG; Apply fPAR2; Apply H; Assumption.

Qed.

(************************************************)
(* Finally the main goal of this package:       *)
(* the congruence of Strong (Late) Bisimilarity *)
(* w.r.t. the ! operator.                       *)
(************************************************)

Inductive BisBANG_S: proc -> proc -> Prop :=
  bisbang_s: (p,q:proc)(StBisim p q)->(r,s:proc)(StBisim r s) -> (BisBANG_S (par r (bang p)) (par s (bang q))).

Lemma BANG_S_L6: (p,q:name->proc)(x:name)
                 (notin x (nu p)) -> (notin x (nu q)) -> (BisBANG_S (p x) (q x)) ->
                 (y:name)(notin y (nu p)) -> (notin y (nu q)) -> (BisBANG_S (p y) (q y)).

Proof.

Intros; Elim (LEM_name x y); Intro.

Rewrite <-H4; Assumption.

Inversion H1.
Elim (proc_exp r x); Elim (proc_exp s x); Elim (proc_exp p0 x); Elim (proc_exp q0 x); Intros.
Inversion_clear H9; Inversion_clear H10; Inversion_clear H11; Inversion_clear H12.
Rewrite H14 in H6; Rewrite H15 in H5; Rewrite H16 in H6; Rewrite H17 in H5.
Cut p=([_:name](par (x3 _) (bang (x1 _)))); [Intro | Apply proc_ext with x; Auto].
Cut q=([_:name](par (x2 _) (bang (x0 _)))); [Intro | Apply proc_ext with x; Auto].
Rewrite H12; Rewrite H18; Apply bisbang_s; Auto.
Rewrite H14 in H7; Rewrite H15 in H7; Apply L6_Light with x; Auto.
Rewrite H12 in H2; Inversion_clear H2; Apply notin_nu; Intros; Cut (notin y (par (x3 z) (bang (x1 z)))); [Intro | Auto]; Inversion_clear H15; Inversion_clear H20; Inversion_clear H21; Assumption.
Rewrite H18 in H3; Inversion_clear H3; Apply notin_nu; Intros; Cut (notin y (par (x2 z) (bang (x0 z)))); [Intro | Auto]; Inversion_clear H15; Inversion_clear H20; Inversion_clear H21; Assumption.
Rewrite H16 in H8; Rewrite H17 in H8; Apply L6_Light with x; Auto.
Rewrite H12 in H2; Inversion_clear H2; Apply notin_nu; Intros; Cut (notin y (par (x3 z) (bang (x1 z)))); [Intro | Auto]; Inversion_clear H15; Inversion_clear H20; Assumption.
Rewrite H18 in H3; Inversion_clear H3; Apply notin_nu; Intros; Cut (notin y (par (x2 z) (bang (x0 z)))); [Intro | Auto]; Inversion_clear H15; Inversion_clear H20; Assumption.
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H10; Auto | Apply notin_bang; Inversion_clear H13; Auto].
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H11; Auto | Apply notin_bang; Inversion_clear H9; Auto].

Qed.

Lemma BisBANG_S_FTR1: (p,q:proc)(BisBANG_S p q) ->
                      (x,y:name)
                      ((p1:proc)(ftrans p (Out x y) p1)->
                       (Ex [q1:proc]((ftrans q (Out x y) q1) /\ (SB_R_SB BisBANG_S p1 q1)))
                      )
                      /\
                      ((q1:proc)(ftrans q (Out x y) q1)->
                       (Ex [p1:proc]((ftrans p (Out x y) p1) /\ (SB_R_SB BisBANG_S p1 q1)))
                      ).

Proof.

Intros; Inversion H; Split; Intros.

Inversion H4.

Inversion_clear H1; Inversion_clear H10.
Elim (H1 (Out x y)); Intros.
Elim (H10 p4 H9); Intros; Inversion_clear H13.
Exists (par x0 (bang q0)); Split; [Apply fPAR1; Assumption | Unfold SB_R_SB; Exists (par p4 (bang p0)); Exists (par x0 (bang q0)); Split; [Apply REF | Split; [Apply bisbang_s; Auto | Apply REF]]].

Elim (BANG_OUT p0 p4 x y H9); Intros; Inversion_clear H10; Inversion_clear H11.
Inversion H0; Inversion_clear H11.
Elim (H15 (Out x y)); Intros.
Elim (H11 x0 H12); Intros; Inversion_clear H18.
Exists (par s (cf_f_act q0 x2 x1)); Split.
Apply fPAR2; Apply BANG_FTR_AUX; Assumption.
Rewrite H10; Unfold SB_R_SB; Exists (par (par r (frepl1 p0 x0 x1)) (bang p0)); Exists (par (par s (frepl1 q0 x2 x1)) (bang q0)); Split.
Apply TRANS with (par r (par (frepl1 p0 x0 x1) (bang p0))); [Apply PAR_S2; [Apply REF | Apply SB_CF_F_ACT; Assumption] | Apply SYM; Apply ASS_PAR].
Split; [Apply bisbang_s; Auto; Apply PAR_S2; [Assumption | Apply F_REPL1; Assumption] | Apply TRANS with (par s (par (frepl1 q0 x2 x1) (bang q0))); [Apply ASS_PAR | Apply PAR_S2; [Apply REF | Apply SYM; Apply SB_CF_F_ACT; Assumption]]].

Inversion H4.

Inversion_clear H1; Inversion_clear H10.
Elim (H1 (Out x y)); Intros.
Elim (H12 p3 H9); Intros; Inversion_clear H13.
Exists (par x0 (bang p0)); Split; [Apply fPAR1; Assumption | Unfold SB_R_SB; Exists (par x0 (bang p0)); Exists (par p3 (bang q0)); Split; [Apply REF | Split; [Apply bisbang_s; Auto | Apply REF]]].

Elim (BANG_OUT q0 p3 x y H9); Intros; Inversion_clear H10; Inversion_clear H11.
Inversion H0; Inversion_clear H11.
Elim (H15 (Out x y)); Intros.
Elim (H17 x0 H12); Intros; Inversion_clear H18.
Exists (par r (cf_f_act p0 x2 x1)); Split.
Apply fPAR2; Apply BANG_FTR_AUX; Assumption.
Rewrite H10; Unfold SB_R_SB; Exists (par (par r (frepl1 p0 x2 x1)) (bang p0)); Exists (par (par s (frepl1 q0 x0 x1)) (bang q0)); Split.
Apply TRANS with (par r (par (frepl1 p0 x2 x1) (bang p0))); [Apply PAR_S2; [Apply REF | Apply SB_CF_F_ACT; Assumption] | Apply SYM; Apply ASS_PAR].
Split; [Apply bisbang_s; Auto; Apply PAR_S2; [Assumption | Apply F_REPL1; Assumption] | Apply TRANS with (par s (par (frepl1 q0 x0 x1) (bang q0))); [Apply ASS_PAR | Apply PAR_S2; [Apply REF | Apply SYM; Apply SB_CF_F_ACT; Assumption]]].

Qed.

Lemma BisBANG_S_IN  : (p,q:proc)(BisBANG_S p q) ->
                      (x:name)
                      ((p1:name->proc)(btrans p (In x) p1)->
                       (Ex [q1:name->proc]((btrans q (In x) q1) /\ (y:name)(SB_R_SB BisBANG_S (p1 y) (q1 y))))
                      )
                      /\
                      ((q1:name->proc)(btrans q (In x) q1)->
                       (Ex [p1:name->proc]((btrans p (In x) p1) /\ (y:name)(SB_R_SB BisBANG_S (p1 y) (q1 y))))
                      ).

Proof.

Intros; Inversion H; Split; Intros.

Inversion H4.

Inversion_clear H1; Inversion_clear H10; Inversion_clear H11.
Elim (H10 x); Intros.
Elim (H11 p4 H9); Intros; Inversion_clear H14.
Exists [_:name](par (x0 _) (bang q0)); Split; [Apply bPAR1; Assumption | Intro; Unfold SB_R_SB; Exists (par (p4 y) (bang p0)); Exists (par (x0 y) (bang q0)); Split; [Apply REF | Split; [Apply bisbang_s; Auto | Apply REF]]].

Elim (BANG_BTR p0 p4 (In x) H9); Intros; Inversion_clear H10; Inversion_clear H11.
Inversion H0; Inversion_clear H11; Inversion_clear H16.
Elim (H11 x); Intros.
Elim (H16 x0 H12); Intros; Inversion_clear H19.
Exists [_:name](par s ((cf_b_act q0 x2 x1) _)); Split.
Apply bPAR2; Apply BANG_BTR_AUX; Assumption.
Rewrite H10; Intro; Unfold SB_R_SB; Exists (par (par r (frepl1 p0 (x0 y) x1)) (bang p0)); Exists (par (par s (frepl1 q0 (x2 y) x1)) (bang q0)); Split.
Apply TRANS with (par r (par (frepl1 p0 (x0 y) x1) (bang p0))); [Apply PAR_S2; [Apply REF | Apply SB_CF_B_ACT; Assumption] | Apply SYM; Apply ASS_PAR].
Split; [Apply bisbang_s; Auto; Apply PAR_S2; [Assumption | Apply F_REPL1; Auto] | Apply TRANS with (par s (par (frepl1 q0 (x2 y) x1) (bang q0))); [Apply ASS_PAR | Apply PAR_S2; [Apply REF | Apply SYM; Apply SB_CF_B_ACT; Assumption]]].

Inversion H4.

Inversion_clear H1; Inversion_clear H10; Inversion_clear H11.
Elim (H10 x); Intros.
Elim (H13 p3 H9); Intros; Inversion_clear H14.
Exists [_:name](par (x0 _) (bang p0)); Split; [Apply bPAR1; Assumption | Intro; Unfold SB_R_SB; Exists (par (x0 y) (bang p0)); Exists (par (p3 y) (bang q0)); Split; [Apply REF | Split; [Apply bisbang_s; Auto | Apply REF]]].

Elim (BANG_BTR q0 p3 (In x) H9); Intros; Inversion_clear H10; Inversion_clear H11.
Inversion H0; Inversion_clear H11; Inversion_clear H16.
Elim (H11 x); Intros.
Elim (H18 x0 H12); Intros; Inversion_clear H19.
Exists [_:name](par r ((cf_b_act p0 x2 x1) _)); Split.
Apply bPAR2; Apply BANG_BTR_AUX; Assumption.
Rewrite H10; Intro; Unfold SB_R_SB; Exists (par (par r (frepl1 p0 (x2 y) x1)) (bang p0)); Exists (par (par s (frepl1 q0 (x0 y) x1)) (bang q0)); Split.
Apply TRANS with (par r (par (frepl1 p0 (x2 y) x1) (bang p0))); [Apply PAR_S2; [Apply REF | Apply SB_CF_B_ACT; Assumption] | Apply SYM; Apply ASS_PAR].
Split; [Apply bisbang_s; Auto; Apply PAR_S2; [Assumption | Apply F_REPL1; Auto] | Apply TRANS with (par s (par (frepl1 q0 (x0 y) x1) (bang q0))); [Apply ASS_PAR | Apply PAR_S2; [Apply REF | Apply SYM; Apply SB_CF_B_ACT; Assumption]]].

Qed.

Lemma BisBANG_S_bOUT  : (p,q:proc)(BisBANG_S p q) ->
                        (x:name)
                        ((p1:name->proc)(btrans p (bOut x) p1)->
                         (Ex [q1:name->proc]((btrans q (bOut x) q1) /\ (y:name)(notin y (nu p1))->(notin y (nu q1))->(SB_R_SB BisBANG_S (p1 y) (q1 y))))
                        )
                        /\
                        ((q1:name->proc)(btrans q (bOut x) q1)->
                         (Ex [p1:name->proc]((btrans p (bOut x) p1) /\ (y:name)(notin y (nu p1))->(notin y (nu q1))->(SB_R_SB BisBANG_S (p1 y) (q1 y))))
                        ).

Proof.

Intros; Inversion H; Split; Intros.

Inversion H4.

Inversion_clear H1; Inversion_clear H10; Inversion_clear H11.
Elim (H12 x); Intros.
Elim (H11 p4 H9); Intros; Inversion_clear H14.
Exists [_:name](par (x0 _) (bang q0)); Split; [Apply bPAR1; Assumption | Intros; Unfold SB_R_SB; Exists (par (p4 y) (bang p0)); Exists (par (x0 y) (bang q0)); Split; [Apply REF | Split; [Apply bisbang_s; Auto; Apply H16; Auto | Apply REF]]].
Inversion_clear H14; Apply notin_nu; Intros; Cut (notin y (par (p4 z) (bang p0))); [Intro | Auto]; Inversion_clear H19; Auto.
Inversion_clear H17; Apply notin_nu; Intros; Cut (notin y (par (x0 z) (bang q0))); [Intro | Auto]; Inversion_clear H19; Auto.

Elim (BANG_BTR p0 p4 (bOut x) H9); Intros; Inversion_clear H10; Inversion_clear H11.
Inversion H0; Inversion_clear H11; Inversion_clear H16.
Elim (H17 x); Intros.
Elim (H16 x0 H12); Intros; Inversion_clear H19.
Exists [_:name](par s ((cf_b_act q0 x2 x1) _)); Split.
Apply bPAR2; Apply BANG_BTR_AUX; Assumption.
Rewrite H10; Intros; Unfold SB_R_SB; Exists (par (par r (frepl1 p0 (x0 y) x1)) (bang p0)); Exists (par (par s (frepl1 q0 (x2 y) x1)) (bang q0)); Split.
Apply TRANS with (par r (par (frepl1 p0 (x0 y) x1) (bang p0))); [Apply PAR_S2; [Apply REF | Apply SB_CF_B_ACT; Assumption] | Apply SYM; Apply ASS_PAR].
Split; [Apply bisbang_s; Auto; Apply PAR_S2; [Assumption | Apply F_REPL1; Auto; Apply H21; [Apply CF_B_ACT_NOTIN with p0 x1; Apply notin_nu; Intros; Inversion_clear H19; Cut (notin y (par r (cf_b_act p0 x0 x1 z))); [Intro | Auto]; Inversion_clear H19; Auto | Apply CF_B_ACT_NOTIN with q0 x1; Apply notin_nu; Intros; Inversion_clear H22; Cut (notin y (par s (cf_b_act q0 x2 x1 z))); [Intro | Auto]; Inversion_clear H22; Auto]] | Apply TRANS with (par s (par (frepl1 q0 (x2 y) x1) (bang q0))); [Apply ASS_PAR | Apply PAR_S2; [Apply REF | Apply SYM; Apply SB_CF_B_ACT; Assumption]]].

Inversion H4.

Inversion_clear H1; Inversion_clear H10; Inversion_clear H11.
Elim (H12 x); Intros.
Elim (H13 p3 H9); Intros; Inversion_clear H14.
Exists [_:name](par (x0 _) (bang p0)); Split; [Apply bPAR1; Assumption | Intros; Unfold SB_R_SB; Exists (par (x0 y) (bang p0));  Exists (par (p3 y) (bang q0)); Split; [Apply REF | Split; [Apply bisbang_s; Auto; Apply H16; Auto | Apply REF]]].
Inversion_clear H14; Apply notin_nu; Intros; Cut (notin y (par (x0 z) (bang p0))); [Intro | Auto]; Inversion_clear H19; Auto.
Inversion_clear H17; Apply notin_nu; Intros; Cut (notin y (par (p3 z) (bang q0))); [Intro | Auto]; Inversion_clear H19; Auto.

Elim (BANG_BTR q0 p3 (bOut x) H9); Intros; Inversion_clear H10; Inversion_clear H11.
Inversion H0; Inversion_clear H11; Inversion_clear H16.
Elim (H17 x); Intros.
Elim (H18 x0 H12); Intros; Inversion_clear H19.
Exists [_:name](par r ((cf_b_act p0 x2 x1) _)); Split.
Apply bPAR2; Apply BANG_BTR_AUX; Assumption.
Rewrite H10; Intros; Unfold SB_R_SB; Exists (par (par r (frepl1 p0 (x2 y) x1)) (bang p0)); Exists (par (par s (frepl1 q0 (x0 y) x1)) (bang q0)); Split.
Apply TRANS with (par r (par (frepl1 p0 (x2 y) x1) (bang p0))); [Apply PAR_S2; [Apply REF | Apply SB_CF_B_ACT; Assumption] | Apply SYM; Apply ASS_PAR].
Split; [Apply bisbang_s; Auto; Apply PAR_S2; [Assumption | Apply F_REPL1; Auto; Apply H21; [Apply CF_B_ACT_NOTIN with p0 x1; Apply notin_nu; Intros; Inversion_clear H19; Cut (notin y (par r (cf_b_act p0 x2 x1 z))); [Intro | Auto]; Inversion_clear H19; Auto | Apply CF_B_ACT_NOTIN with q0 x1; Apply notin_nu; Intros; Inversion_clear H22; Cut (notin y (par s (cf_b_act q0 x0 x1 z))); [Intro | Auto]; Inversion_clear H22; Auto]] | Apply TRANS with (par s (par (frepl1 q0 (x0 y) x1) (bang q0))); [Apply ASS_PAR | Apply PAR_S2; [Apply REF | Apply SYM; Apply SB_CF_B_ACT; Assumption]]].

Qed.

Lemma BisBANG_S_FTR2: (p,q:proc)(BisBANG_S p q) ->
                      ((p1:proc)(ftrans p tau p1)->
                       (Ex [q1:proc](ftrans q tau q1) /\
                           ((Ex [p1':proc](Ex [q1':proc](StBisim p1 p1')
                                                    /\ (BisBANG_S p1' q1')
                                                    /\ (StBisim q1' q1)))
                            \/
                            (Ex [p1':name->proc](Ex [q1':name->proc](StBisim p1 (nu p1'))
                                                                 /\ ((z:name)
                                                                     (notin z (nu p1'))->
                                                                     (notin z (nu q1'))->
                                                                     (BisBANG_S (p1' z) (q1' z))
                                                                    )
                                                                 /\ (StBisim (nu q1') q1)))
                           )
                       )
                      )
                      /\
                      ((q1:proc)(ftrans q tau q1)->
                       (Ex [p1:proc](ftrans p tau p1) /\
                           ((Ex [p1':proc](Ex [q1':proc](StBisim p1 p1')
                                                    /\ (BisBANG_S p1' q1')
                                                    /\ (StBisim q1' q1)))
                            \/
                            (Ex [p1':name->proc](Ex [q1':name->proc](StBisim p1 (nu p1'))
                                                                 /\ ((z:name)
                                                                     (notin z (nu p1'))->
                                                                     (notin z (nu q1'))->
                                                                     (BisBANG_S (p1' z) (q1' z))
                                                                    )
                                                                 /\ (StBisim (nu q1') q1)))
                           )
                       )
                      ).

Proof.

Intros; Inversion H; Split; Intros.

Inversion H4.

Inversion_clear H1; Inversion_clear H10.
Elim (H1 tau); Intros.
Elim (H10 p4 H9); Intros; Inversion_clear H13.
Exists (par x (bang q0)); Split; [Apply fPAR1; Assumption | Left; Exists (par p4 (bang p0)); Exists (par x (bang q0)); Split; [Apply REF | Split; [Apply bisbang_s; Auto | Apply REF]]].

Elim (BANG_TAU p0 p4 H9); Intros; Inversion_clear H10; Inversion_clear H11; Inversion_clear H10.
Inversion H0; Inversion_clear H10.
Elim (H15 tau); Intros.
Elim (H10 x H12); Intros; Inversion_clear H18.
Exists (par s (cf_f_act q0 x1 x0)); Split; [Apply fPAR2; Apply BANG_FTR_AUX; Assumption | Rewrite H11; Left; Exists (par (par r (frepl1 p0 x x0)) (bang p0)); Exists (par (par s (frepl1 q0 x1 x0)) (bang q0)); Split].
Apply TRANS with (par r (par (frepl1 p0 x x0) (bang p0))); [Apply PAR_S2; [Apply REF | Apply SB_CF_F_ACT] | Apply SYM; Apply ASS_PAR].
Split; [Apply bisbang_s; [Assumption | Apply PAR_S2; [Assumption | Apply F_REPL1; Assumption]] | Apply TRANS with (par s (par (frepl1 q0 x1 x0) (bang q0))); [Apply ASS_PAR | Apply PAR_S2; [Apply REF | Apply SYM; Apply SB_CF_F_ACT]]].

Inversion_clear H11; Inversion_clear H10; Inversion_clear H11; Inversion_clear H10; Inversion_clear H11; Inversion_clear H12.
Inversion H0; Inversion_clear H12; Inversion_clear H17.
Elim (H16 (Out x1 x2)); Elim (H12 x1); Intros.
Elim (H17 x0 H13); Elim (H20 x H11); Intros; Inversion_clear H22; Inversion_clear H23.
Exists (par s (cf_tau1 q0 x5 x6 x2 x3 x4)); Split; [Apply fPAR2; Apply BANG_TAU_AUX1 with x1; Assumption | Rewrite H10; Left; Exists (par (par r (frepl2 p0 x x0 x2 x3 x4)) (bang p0)); Exists (par (par s (frepl2 q0 x5 x6 x2 x3 x4)) (bang q0)); Split].
Apply TRANS with (par r (par (frepl2 p0 x x0 x2 x3 x4) (bang p0))); [Apply PAR_S2; [Apply REF | Apply SB_CF_TAU1] | Apply SYM; Apply ASS_PAR].
Split; [Apply bisbang_s; [Assumption | Apply PAR_S2; [Assumption | Apply F_REPL2; Auto]] | Apply TRANS with (par s (par (frepl2 q0 x5 x6 x2 x3 x4) (bang q0))); [Apply ASS_PAR | Apply PAR_S2; [Apply REF | Apply SYM; Apply SB_CF_TAU1]]].

Inversion_clear H11; Inversion_clear H10; Inversion_clear H11; Inversion_clear H10; Inversion_clear H11; Inversion_clear H10; Inversion_clear H12.
Inversion H0; Inversion_clear H12; Inversion_clear H17.
Elim (H16 (Out x1 x2)); Elim (H12 x1); Intros.
Elim (H17 x0 H13); Elim (H20 x H10); Intros; Inversion_clear H22; Inversion_clear H23.
Exists (par s (cf_tau2 q0 x5 x6 x2 x3 x4)); Split; [Apply fPAR2; Apply BANG_TAU_AUX2 with x1; Assumption | Rewrite H11; Left; Exists (par (par r (frepl3 p0 x x0 x2 x3 x4)) (bang p0)); Exists (par (par s (frepl3 q0 x5 x6 x2 x3 x4)) (bang q0)); Split].
Apply TRANS with (par r (par (frepl3 p0 x x0 x2 x3 x4) (bang p0))); [Apply PAR_S2; [Apply REF | Apply SB_CF_TAU2] | Apply SYM; Apply ASS_PAR].
Split; [Apply bisbang_s; [Assumption | Apply PAR_S2; [Assumption | Apply F_REPL3; Auto]] | Apply TRANS with (par s (par (frepl3 q0 x5 x6 x2 x3 x4) (bang q0))); [Apply ASS_PAR | Apply PAR_S2; [Apply REF | Apply SYM; Apply SB_CF_TAU2]]].

Inversion_clear H11; Inversion_clear H10; Inversion_clear H11; Inversion_clear H10; Inversion_clear H11; Inversion_clear H10; Inversion_clear H12.
Inversion H0; Inversion_clear H12; Inversion_clear H17.
Elim (H12 x1); Elim (H18 x1); Intros.
Elim (H17 x0 H13); Elim (H20 x H10); Intros; Inversion_clear H22; Inversion_clear H23.
Exists (par s (cf_tau3 q0 x4 x5 x2 x3)); Split.
Apply fPAR2; Apply BANG_TAU_AUX3 with x1; Assumption.
Rewrite H11; Left; Exists (par (par r (frepl4 p0 x x0 x2 x3)) (bang p0)); Exists (par (par s (frepl4 q0 x4 x5 x2 x3)) (bang q0)); Split.
Apply TRANS with (par r (par (frepl4 p0 x x0 x2 x3) (bang p0))); [Apply PAR_S2; [Apply REF | Apply SB_CF_TAU3; Auto] | Apply SYM; Apply ASS_PAR].
Split; [Apply bisbang_s; [Assumption | Apply PAR_S2; [Assumption | Apply F_REPL4; Auto]] | Apply TRANS with (par s (par (frepl4 q0 x4 x5 x2 x3) (bang q0))); [Apply ASS_PAR | Apply PAR_S2; [Apply REF | Apply SYM; Apply SB_CF_TAU3; Assumption]]].

Inversion_clear H11; Inversion_clear H10; Inversion_clear H11; Inversion_clear H10; Inversion_clear H11; Inversion_clear H10; Inversion_clear H12.
Inversion H0; Inversion_clear H12; Inversion_clear H17.
Elim (H12 x1); Elim (H18 x1); Intros.
Elim (H17 x0 H13); Elim (H20 x H10); Intros; Inversion_clear H22; Inversion_clear H23.
Exists (par s (cf_tau3 q0 x5 x4 x2 x3)); Split.
Apply fPAR2; Apply BANG_TAU_AUX4 with x1; Assumption.
Rewrite H11; Left; Exists (par (par r (frepl4 p0 x0 x x2 x3)) (bang p0)); Exists (par (par s (frepl4 q0 x5 x4 x2 x3)) (bang q0)); Split.
Apply TRANS with (par r (par (frepl4 p0 x0 x x2 x3) (bang p0))); [Apply PAR_S2; [Apply REF | Apply SB_CF_TAU3; Auto] | Apply SYM; Apply ASS_PAR].
Split; [Apply bisbang_s; [Assumption | Apply PAR_S2; [Assumption | Apply F_REPL4; Auto]] | Apply TRANS with (par s (par (frepl4 q0 x5 x4 x2 x3) (bang q0))); [Apply ASS_PAR | Apply PAR_S2; [Apply REF | Apply SYM; Apply SB_CF_TAU3; Assumption]]].

Elim (BANG_OUT p0 q2 x y H10); Intros; Inversion_clear H11; Inversion_clear H12; Inversion H0; Inversion H1; Inversion_clear H12; Inversion_clear H16; Inversion_clear H21.
Elim (H19 (Out x y)); Elim (H16 x); Intros.
Elim (H21 q1 H7); Elim (H24 x0 H13); Intros; Inversion_clear H26; Inversion_clear H27.
Exists (par (x3 y) (cf_f_act q0 x2 x1)); Split.
Apply COM1 with x; [Assumption | Apply BANG_FTR_AUX; Assumption].
Rewrite H11; Left; Exists (par (par (q1 y) (frepl1 p0 x0 x1)) (bang p0)); Exists (par (par (x3 y) (frepl1 q0 x2 x1)) (bang q0)); Split.
Apply TRANS with (par (q1 y) (par (frepl1 p0 x0 x1) (bang p0))); [Apply PAR_S2; [Apply REF | Apply SB_CF_F_ACT; Assumption] | Apply SYM; Apply ASS_PAR].
Split; [Apply bisbang_s; [Assumption | Apply PAR_S2; [Auto | Apply F_REPL1; Assumption]] | Apply TRANS with (par (x3 y) (par (frepl1 q0 x2 x1) (bang q0))); [Apply ASS_PAR | Apply PAR_S2; [Apply REF | Apply SYM; Apply SB_CF_F_ACT; Assumption]]].

Elim (BANG_BTR p0 q2 (In x) H10); Intros; Inversion_clear H11; Inversion_clear H12; Inversion H0; Inversion H1; Inversion_clear H12; Inversion_clear H16; Inversion_clear H20.
Elim (H12 (Out x y)); Elim (H16 x); Intros.
Elim (H20 x0 H13); Elim (H24 q1 H7); Intros; Inversion_clear H26; Inversion_clear H27.
Exists (par x2 ((cf_b_act q0 x3 x1) y)); Split.
Apply COM2 with x; [Assumption | Apply BANG_BTR_AUX; Assumption].
Rewrite H11; Left; Exists (par (par q1 (frepl1 p0 (x0 y) x1)) (bang p0)); Exists (par (par x2 (frepl1 q0 (x3 y) x1)) (bang q0)); Split.
Apply TRANS with (par q1 (par (frepl1 p0 (x0 y) x1) (bang p0))); [Apply PAR_S2; [Apply REF | Apply SB_CF_B_ACT; Assumption] | Apply SYM; Apply ASS_PAR].
Split; [Apply bisbang_s; [Assumption | Apply PAR_S2; [Auto | Apply F_REPL1; Auto]] | Apply TRANS with (par x2 (par (frepl1 q0 (x3 y) x1) (bang q0))); [Apply ASS_PAR | Apply PAR_S2; [Apply REF | Apply SYM; Apply SB_CF_B_ACT; Assumption]]].

Elim (BANG_BTR p0 q2 (bOut x) H10); Intros; Inversion_clear H11; Inversion_clear H12; Inversion H0; Inversion H1; Inversion_clear H12; Inversion_clear H16; Inversion_clear H20; Inversion_clear H21.
Elim (H20 x); Elim (H22 x); Intros.
Elim (H21 x0 H13); Elim (H25 q1 H7); Intros; Inversion_clear H27; Inversion_clear H28.
Exists (nu [_:name](par (x2 _) ((cf_b_act q0 x3 x1) _))); Split.
Apply CLOSE1 with x; [Assumption | Apply BANG_BTR_AUX; Assumption].
Rewrite H11; Right.
Exists [_:name](par (par (q1 _) (frepl1 p0 (x0 _) x1)) (bang p0)); Exists [_:name](par (par (x2 _) (frepl1 q0 (x3 _) x1)) (bang q0)); Split.
Apply NU_S; Intros; Apply TRANS with (par (q1 x4) (par (frepl1 p0 (x0 x4) x1) (bang p0))); [Apply PAR_S2; [Apply REF | Apply SB_CF_B_ACT; Assumption] | Apply SYM; Apply ASS_PAR].
Split; [Intros; Apply bisbang_s; [Assumption | Apply PAR_S2; [Auto | Apply F_REPL1; Auto; Apply H31; [Apply notin_nu; Intros; Apply FREPL1_NOTIN with p0 x1; Inversion_clear H28; Cut (notin z (par (par (q1 z0) (frepl1 p0 (x0 z0) x1)) (bang p0))); [Intro | Auto]; Inversion_clear H28; Inversion_clear H35; Assumption | Apply notin_nu; Intros; Apply FREPL1_NOTIN with q0 x1; Inversion_clear H32; Cut (notin z (par (par (x2 z0) (frepl1 q0 (x3 z0) x1)) (bang q0))); [Intro | Auto]; Inversion_clear H32; Inversion_clear H35; Assumption]]] | Apply NU_S; Intros; Apply TRANS with (par (x2 x4) (par (frepl1 q0 (x3 x4) x1) (bang q0))); [Apply ASS_PAR | Apply PAR_S2; [Apply REF | Apply SYM; Apply SB_CF_B_ACT; Assumption]]].

Elim (BANG_BTR p0 q2 (In x) H10); Intros; Inversion_clear H11; Inversion_clear H12; Inversion H0; Inversion H1; Inversion_clear H12; Inversion_clear H16; Inversion_clear H20; Inversion_clear H21.
Elim (H16 x); Elim (H23 x); Intros.
Elim (H21 q1 H7); Elim (H25 x0 H13); Intros; Inversion_clear H27; Inversion_clear H28.
Exists (nu [_:name](par (x3 _) ((cf_b_act q0 x2 x1) _))); Split.
Apply CLOSE2 with x; [Assumption | Apply BANG_BTR_AUX; Assumption].
Rewrite H11; Right.
Exists [_:name](par (par (q1 _) (frepl1 p0 (x0 _) x1)) (bang p0)); Exists [_:name](par (par (x3 _) (frepl1 q0 (x2 _) x1)) (bang q0)); Split.
Apply NU_S; Intros; Apply TRANS with (par (q1 x4) (par (frepl1 p0 (x0 x4) x1) (bang p0))); [Apply PAR_S2; [Apply REF | Apply SB_CF_B_ACT; Assumption] | Apply SYM; Apply ASS_PAR].
Split; [Intros; Apply bisbang_s; [Assumption | Apply PAR_S2; [Apply H31; [Apply notin_nu; Intros; Inversion_clear H28; Cut (notin z (par (par (q1 z0) (frepl1 p0 (x0 z0) x1)) (bang p0))); [Intro | Auto]; Inversion_clear H28; Inversion_clear H35; Assumption | Apply notin_nu; Intros; Inversion_clear H32; Cut (notin z (par (par (x3 z0) (frepl1 q0 (x2 z0) x1)) (bang q0))); [Intro | Auto]; Inversion_clear H32; Inversion_clear H35; Assumption] | Apply F_REPL1; Auto]] | Apply NU_S; Intros; Apply TRANS with (par (x3 x4) (par (frepl1 q0 (x2 x4) x1) (bang q0))); [Apply ASS_PAR | Apply PAR_S2; [Apply REF | Apply SYM; Apply SB_CF_B_ACT; Assumption]]].

Inversion H4.

Inversion_clear H1; Inversion_clear H10.
Elim (H1 tau); Intros.
Elim (H12 p3 H9); Intros; Inversion_clear H13.
Exists (par x (bang p0)); Split; [Apply fPAR1; Assumption | Left; Exists (par x (bang p0)); Exists (par p3 (bang q0)); Split; [Apply REF | Split; [Apply bisbang_s; Auto | Apply REF]]].

Elim (BANG_TAU q0 p3 H9); Intros; Inversion_clear H10; Inversion_clear H11; Inversion_clear H10.
Inversion H0; Inversion_clear H10.
Elim (H15 tau); Intros.
Elim (H17 x H12); Intros; Inversion_clear H18.
Exists (par r (cf_f_act p0 x1 x0)); Split; [Apply fPAR2; Apply BANG_FTR_AUX; Assumption | Rewrite H11; Left; Exists (par (par r (frepl1 p0 x1 x0)) (bang p0)); Exists (par (par s (frepl1 q0 x x0)) (bang q0)); Split].
Apply TRANS with (par r (par (frepl1 p0 x1 x0) (bang p0))); [Apply PAR_S2; [Apply REF | Apply SB_CF_F_ACT] | Apply SYM; Apply ASS_PAR].
Split; [Apply bisbang_s; [Assumption | Apply PAR_S2; [Assumption | Apply F_REPL1; Assumption]] | Apply TRANS with (par s (par (frepl1 q0 x x0) (bang q0))); [Apply ASS_PAR | Apply PAR_S2; [Apply REF | Apply SYM; Apply SB_CF_F_ACT]]].

Inversion_clear H11; Inversion_clear H10; Inversion_clear H11; Inversion_clear H10; Inversion_clear H11; Inversion_clear H12.
Inversion H0; Inversion_clear H12; Inversion_clear H17.
Elim (H16 (Out x1 x2)); Elim (H12 x1); Intros.
Elim (H19 x0 H13); Elim (H21 x H11); Intros; Inversion_clear H22; Inversion_clear H23.
Exists (par r (cf_tau1 p0 x5 x6 x2 x3 x4)); Split; [Apply fPAR2; Apply BANG_TAU_AUX1 with x1; Assumption | Rewrite H10; Left; Exists (par (par r (frepl2 p0 x5 x6 x2 x3 x4)) (bang p0)); Exists (par (par s (frepl2 q0 x x0 x2 x3 x4)) (bang q0)); Split].
Apply TRANS with (par r (par (frepl2 p0 x5 x6 x2 x3 x4) (bang p0))); [Apply PAR_S2; [Apply REF | Apply SB_CF_TAU1] | Apply SYM; Apply ASS_PAR].
Split; [Apply bisbang_s; [Assumption | Apply PAR_S2; [Assumption | Apply F_REPL2; Auto]] | Apply TRANS with (par s (par (frepl2 q0 x x0 x2 x3 x4) (bang q0))); [Apply ASS_PAR | Apply PAR_S2; [Apply REF | Apply SYM; Apply SB_CF_TAU1]]].

Inversion_clear H11; Inversion_clear H10; Inversion_clear H11; Inversion_clear H10; Inversion_clear H11; Inversion_clear H10; Inversion_clear H12.
Inversion H0; Inversion_clear H12; Inversion_clear H17.
Elim (H16 (Out x1 x2)); Elim (H12 x1); Intros.
Elim (H19 x0 H13); Elim (H21 x H10); Intros; Inversion_clear H22; Inversion_clear H23.
Exists (par r (cf_tau2 p0 x5 x6 x2 x3 x4)); Split; [Apply fPAR2; Apply BANG_TAU_AUX2 with x1; Assumption | Rewrite H11; Left; Exists (par (par r (frepl3 p0 x5 x6 x2 x3 x4)) (bang p0)); Exists (par (par s (frepl3 q0 x x0 x2 x3 x4)) (bang q0)); Split].
Apply TRANS with (par r (par (frepl3 p0 x5 x6 x2 x3 x4) (bang p0))); [Apply PAR_S2; [Apply REF | Apply SB_CF_TAU2] | Apply SYM; Apply ASS_PAR].
Split; [Apply bisbang_s; [Assumption | Apply PAR_S2; [Assumption | Apply F_REPL3; Auto]] | Apply TRANS with (par s (par (frepl3 q0 x x0 x2 x3 x4) (bang q0))); [Apply ASS_PAR | Apply PAR_S2; [Apply REF | Apply SYM; Apply SB_CF_TAU2]]].

Inversion_clear H11; Inversion_clear H10; Inversion_clear H11; Inversion_clear H10; Inversion_clear H11; Inversion_clear H10; Inversion_clear H12.
Inversion H0; Inversion_clear H12; Inversion_clear H17.
Elim (H12 x1); Elim (H18 x1); Intros.
Elim (H19 x0 H13); Elim (H21 x H10); Intros; Inversion_clear H22; Inversion_clear H23.
Exists (par r (cf_tau3 p0 x4 x5 x2 x3)); Split.
Apply fPAR2; Apply BANG_TAU_AUX3 with x1; Assumption.
Rewrite H11; Left; Exists (par (par r (frepl4 p0 x4 x5 x2 x3)) (bang p0)); Exists (par (par s (frepl4 q0 x x0 x2 x3)) (bang q0)); Split.
Apply TRANS with (par r (par (frepl4 p0 x4 x5 x2 x3) (bang p0))); [Apply PAR_S2; [Apply REF | Apply SB_CF_TAU3; Auto] | Apply SYM; Apply ASS_PAR].
Split; [Apply bisbang_s; [Assumption | Apply PAR_S2; [Assumption | Apply F_REPL4; Auto]] | Apply TRANS with (par s (par (frepl4 q0 x x0 x2 x3) (bang q0))); [Apply ASS_PAR | Apply PAR_S2; [Apply REF | Apply SYM; Apply SB_CF_TAU3; Assumption]]].

Inversion_clear H11; Inversion_clear H10; Inversion_clear H11; Inversion_clear H10; Inversion_clear H11; Inversion_clear H10; Inversion_clear H12.
Inversion H0; Inversion_clear H12; Inversion_clear H17.
Elim (H12 x1); Elim (H18 x1); Intros.
Elim (H19 x0 H13); Elim (H21 x H10); Intros; Inversion_clear H22; Inversion_clear H23.
Exists (par r (cf_tau3 p0 x5 x4 x2 x3)); Split.
Apply fPAR2; Apply BANG_TAU_AUX4 with x1; Assumption.
Rewrite H11; Left; Exists (par (par r (frepl4 p0 x5 x4 x2 x3)) (bang p0)); Exists (par (par s (frepl4 q0 x0 x x2 x3)) (bang q0)); Split.
Apply TRANS with (par r (par (frepl4 p0 x5 x4 x2 x3) (bang p0))); [Apply PAR_S2; [Apply REF | Apply SB_CF_TAU3; Auto] | Apply SYM; Apply ASS_PAR].
Split; [Apply bisbang_s; [Assumption | Apply PAR_S2; [Assumption | Apply F_REPL4; Auto]] | Apply TRANS with (par s (par (frepl4 q0 x0 x x2 x3) (bang q0))); [Apply ASS_PAR | Apply PAR_S2; [Apply REF | Apply SYM; Apply SB_CF_TAU3; Assumption]]].

Elim (BANG_OUT q0 q2 x y H10); Intros; Inversion_clear H11; Inversion_clear H12; Inversion H0; Inversion H1; Inversion_clear H12; Inversion_clear H16; Inversion_clear H21.
Elim (H19 (Out x y)); Elim (H16 x); Intros.
Elim (H23 q3 H7); Elim (H25 x0 H13); Intros; Inversion_clear H26; Inversion_clear H27.
Exists (par (x3 y) (cf_f_act p0 x2 x1)); Split.
Apply COM1 with x; [Assumption | Apply BANG_FTR_AUX; Assumption].
Rewrite H11; Left; Exists (par (par (x3 y) (frepl1 p0 x2 x1)) (bang p0)); Exists (par (par (q3 y) (frepl1 q0 x0 x1)) (bang q0)); Split.
Apply TRANS with (par (x3 y) (par (frepl1 p0 x2 x1) (bang p0))); [Apply PAR_S2; [Apply REF | Apply SB_CF_F_ACT; Assumption] | Apply SYM; Apply ASS_PAR].
Split; [Apply bisbang_s; [Assumption | Apply PAR_S2; [Auto | Apply F_REPL1; Assumption]] | Apply TRANS with (par (q3 y) (par (frepl1 q0 x0 x1) (bang q0))); [Apply ASS_PAR | Apply PAR_S2; [Apply REF | Apply SYM; Apply SB_CF_F_ACT; Assumption]]].

Elim (BANG_BTR q0 q3 (In x) H10); Intros; Inversion_clear H11; Inversion_clear H12; Inversion H0; Inversion H1; Inversion_clear H12; Inversion_clear H16; Inversion_clear H20.
Elim (H12 (Out x y)); Elim (H16 x); Intros.
Elim (H23 x0 H13); Elim (H25 q2 H7); Intros; Inversion_clear H26; Inversion_clear H27.
Exists (par x2 ((cf_b_act p0 x3 x1) y)); Split.
Apply COM2 with x; [Assumption | Apply BANG_BTR_AUX; Assumption].
Rewrite H11; Left; Exists (par (par x2 (frepl1 p0 (x3 y) x1)) (bang p0)); Exists (par (par q2 (frepl1 q0 (x0 y) x1)) (bang q0)); Split.
Apply TRANS with (par x2 (par (frepl1 p0 (x3 y) x1) (bang p0))); [Apply PAR_S2; [Apply REF | Apply SB_CF_B_ACT; Assumption] | Apply SYM; Apply ASS_PAR].
Split; [Apply bisbang_s; [Assumption | Apply PAR_S2; [Auto | Apply F_REPL1; Auto]] | Apply TRANS with (par q2 (par (frepl1 q0 (x0 y) x1) (bang q0))); [Apply ASS_PAR | Apply PAR_S2; [Apply REF | Apply SYM; Apply SB_CF_B_ACT; Assumption]]].

Elim (BANG_BTR q0 q3 (bOut x) H10); Intros; Inversion_clear H11; Inversion_clear H12; Inversion H0; Inversion H1; Inversion_clear H12; Inversion_clear H16; Inversion_clear H20; Inversion_clear H21.
Elim (H20 x); Elim (H22 x); Intros.
Elim (H24 x0 H13); Elim (H26 q2 H7); Intros; Inversion_clear H27; Inversion_clear H28.
Exists (nu [_:name](par (x2 _) ((cf_b_act p0 x3 x1) _))); Split.
Apply CLOSE1 with x; [Assumption | Apply BANG_BTR_AUX; Assumption].
Rewrite H11; Right.
Exists [_:name](par (par (x2 _) (frepl1 p0 (x3 _) x1)) (bang p0)); Exists [_:name](par (par (q2 _) (frepl1 q0 (x0 _) x1)) (bang q0)); Split.
Apply NU_S; Intros; Apply TRANS with (par (x2 x4) (par (frepl1 p0 (x3 x4) x1) (bang p0))); [Apply PAR_S2; [Apply REF | Apply SB_CF_B_ACT; Assumption] | Apply SYM; Apply ASS_PAR].
Split; [Intros; Apply bisbang_s; [Assumption | Apply PAR_S2; [Auto | Apply F_REPL1; Auto; Apply H31; [Apply notin_nu; Intros; Apply FREPL1_NOTIN with p0 x1; Inversion_clear H28; Cut (notin z (par (par (x2 z0) (frepl1 p0 (x3 z0) x1)) (bang p0))); [Intro | Auto]; Inversion_clear H28; Inversion_clear H35; Assumption | Apply notin_nu; Intros; Apply FREPL1_NOTIN with q0 x1; Inversion_clear H32; Cut (notin z (par (par (q2 z0) (frepl1 q0 (x0 z0) x1)) (bang q0))); [Intro | Auto]; Inversion_clear H32; Inversion_clear H35; Assumption]]] | Apply NU_S; Intros; Apply TRANS with (par (q2 x4) (par (frepl1 q0 (x0 x4) x1) (bang q0))); [Apply ASS_PAR | Apply PAR_S2; [Apply REF | Apply SYM; Apply SB_CF_B_ACT; Assumption]]].

Elim (BANG_BTR q0 q3 (In x) H10); Intros; Inversion_clear H11; Inversion_clear H12; Inversion H0; Inversion H1; Inversion_clear H12; Inversion_clear H16; Inversion_clear H20; Inversion_clear H21.
Elim (H16 x); Elim (H23 x); Intros.
Elim (H24 q2 H7); Elim (H26 x0 H13); Intros; Inversion_clear H27; Inversion_clear H28.
Exists (nu [_:name](par (x3 _) ((cf_b_act p0 x2 x1) _))); Split.
Apply CLOSE2 with x; [Assumption | Apply BANG_BTR_AUX; Assumption].
Rewrite H11; Right.
Exists [_:name](par (par (x3 _) (frepl1 p0 (x2 _) x1)) (bang p0)); Exists [_:name](par (par (q2 _) (frepl1 q0 (x0 _) x1)) (bang q0)); Split.
Apply NU_S; Intros; Apply TRANS with (par (x3 x4) (par (frepl1 p0 (x2 x4) x1) (bang p0))); [Apply PAR_S2; [Apply REF | Apply SB_CF_B_ACT; Assumption] | Apply SYM; Apply ASS_PAR].
Split; [Intros; Apply bisbang_s; [Assumption | Apply PAR_S2; [Apply H31; [Apply notin_nu; Intros; Inversion_clear H28; Cut (notin z (par (par (x3 z0) (frepl1 p0 (x2 z0) x1)) (bang p0))); [Intro | Auto]; Inversion_clear H28; Inversion_clear H35; Assumption | Apply notin_nu; Intros; Inversion_clear H32; Cut (notin z (par (par (q2 z0) (frepl1 q0 (x0 z0) x1)) (bang q0))); [Intro | Auto]; Inversion_clear H32; Inversion_clear H35; Assumption] | Apply F_REPL1; Auto]] | Apply NU_S; Intros; Apply TRANS with (par (q2 x4) (par (frepl1 q0 (x0 x4) x1) (bang q0))); [Apply ASS_PAR | Apply PAR_S2; [Apply REF | Apply SYM; Apply SB_CF_B_ACT; Assumption]]].

Qed.

Lemma BisBANG_S_UPTO: (UpTo BisBANG_S).

Proof.

Unfold UpTo; Intros.

Split; Intros.

Rewrite H2 in H; Rewrite H3 in H; Apply BANG_S_L6 with x; Auto.

Split; Intros.

Apply BisBANG_S_FTR1; Auto.

Split; Intros.

Apply BisBANG_S_IN; Assumption.

Split; Intros.

Apply BisBANG_S_bOUT; Assumption.

Apply BisBANG_S_FTR2; Assumption.

Qed.

Lemma BANG_S: (p,q:proc)(StBisim p q)->(StBisim (bang p) (bang q)).

Proof.

Intros; Apply Completeness; Apply Co_Ind with (SBUPTO BisBANG_S); [Apply UPTO_SB'; Exact BisBANG_S_UPTO | Apply sbupto with O; Simpl; Unfold SB_R_SB; Exists (par nil (bang p)); Exists (par nil (bang q)); Split; [Apply SYM; Apply TRANS with (par (bang p) nil); [Apply COMM_PAR | Apply ID_PAR] | Split; [Apply bisbang_s; [Assumption | Apply REF] | Apply TRANS with (par (bang q) nil); [Apply COMM_PAR | Apply ID_PAR]]]].

Qed.

End structural_congruence3.
